package jp.co.sds_corp.fxchart.graph;

import java.io.BufferedReader;
import java.io.FileReader;
import java.io.IOException;
import java.math.BigDecimal;
import java.text.DecimalFormat;
import java.text.ParseException;
import java.util.Iterator;
import java.util.Map;
import java.util.TreeMap;

import jp.co.sds_corp.fxutil.CcyPair;
import jp.co.sds_corp.util.common.CmnDate;
import jp.co.sds_corp.util.common.CmnString;

/**
* ファイルからレートを読み込むIRateSourceの実装です。
* その他のレートソースクラスの基底クラスでもあります。
*/
public class RateFileSource implements IRateSource
{
	String	dir;
	CcyPair	pair;
	String	term;
	String	span;
	double  dMax = -1.0;
	double  dMin = 999999999999999.0;

	/**
	* レートソースを生成します。
	* @param	pair	通貨ペア
	* @param	term	期間コード
	* @param	span	チャート種別
	*/
	public RateFileSource( CcyPair pair, String term, String span ) {
		this.pair = pair;
		this.term = term;
		this.span = span;
	}

	/**
	* ファイルの場所を設定します。
	* @param	dir		ファイルの場所
	*/
	public void setDir( String dir ) {
		this.dir = dir;
	}

	/**
	* 指定された期間内の最小レートを取得します。
	* getData()した後にのみ有効な値を返します。
	*/
	public double getMin() {
		return dMin;
	}

	/**
	* 指定された期間内の最大レートを取得します。
	* getData()した後にのみ有効な値を返します。
	*/
	public double getMax() {
		return dMax;
	}

	/**
	* レートソースからレートを読み込み、指定された種別にあった形の
	* RateHistRosokuクラスにデータを集約して返します。
	* @param	fr		期間from
	* @param	to		期間to
	* @param	span	チャート種別
	* @return	RateHistRosokuが格納されたMap
	*/
	public Map<String,RateHistRosoku> getData( CmnDate fr, CmnDate to, String span )
		throws Exception
	{
		Map<String,RateHistRosoku>	mapRate = getData( fr, to );
		if ( span.equals( "TICK" )) {
			return mapRate;
		}

		TreeMap<String,RateHistRosoku>	mapHist = new TreeMap<String,RateHistRosoku>();
		Iterator<String> itr = mapRate.keySet().iterator();
		while ( itr.hasNext()) {
			String hkey = (String )itr.next();
			RateHistRosoku rate = (RateHistRosoku )mapRate.get( hkey );

			CmnDate	dt = rate.getStartDate();
			String key = getHistoryKey( span, rate );

			RateHistRosoku hist = (RateHistRosoku )mapHist.get( key );
			if ( hist == null ) {
				hist = new RateHistRosoku();
				hist.setStartDate( rate.getStartDate() );
				hist.setEndDate( rate.getEndDate() );
				hist.setStart( rate.getStart() );
				hist.setEnd( rate.getEnd() );
				hist.setHigh( rate.getHigh() );
				hist.setLow( rate.getLow() );
			} else {
				if ( hist.high.doubleValue() < rate.high.doubleValue() ) {
					hist.high = rate.high;
				}
				if ( hist.low.doubleValue() > rate.low.doubleValue() ) {
					hist.low = rate.low;
				}
				if ( hist.getEndDate().before( dt ) ) {
					hist.setEndDate( dt );
					hist.end = rate.end;
				}
			}
			mapHist.put( key, hist );
		}
		return mapHist;
	}

	/**
	* レートを取得するファイル名を返します。
	* @param	date	受信日時
	*/
	public String getFileName( CmnDate date ) {
		String ret = dir + "/" + date.format( "yyyyMM" )
					+ "/rate." + date.format( "yyyyMMdd" );
		return ret;
	}

	/**
	* ファイルからレートを読み込み、そのままの形でRateHistRosokuに
	* 設定して返します。
	* @param	fr		期間from
	* @param	to		期間to
	* @return	RateHistRosokuが格納されたMap
	*/
	public Map<String,RateHistRosoku> getData( CmnDate fr, CmnDate to ) throws Exception {
		dMax = -1.0;
		dMin = 999999999999999.0;
		TreeMap<String,RateHistRosoku> mapHist = new TreeMap<String,RateHistRosoku>();
		CmnDate fdt = (CmnDate )fr.clone();
		fdt.clearTime();
		fdt.set( CmnDate.HOUR_OF_DAY, 0 );
		for ( ;; ) {
			BufferedReader in = null;
			try {
				FileReader fin = new FileReader( getFileName( fdt ) );
				in = new BufferedReader( fin );
			} catch ( IOException e ) {
			}

			if ( in != null ) {
				RateHistRosoku pre = null;
				for ( ;; ) {
					String line = in.readLine();
					if ( line == null ) {
						break;
					}
					RateHistRosoku hist = parseRate( line );
					if ( hist == null ) {
						continue;
					}
					if ( hist.getStartDate().before( fr )) {
						pre = hist;
						continue;
					}

					String key = null;
					if ( pre != null ) {
						key = getRateKey( pre );
						mapHist.put( key, pre );
						pre = null;
					}

					key = getRateKey( hist );
					mapHist.put( key, hist );
					if ( dMax < hist.getHigh().doubleValue() ) {
						dMax = hist.getHigh().doubleValue();
					}
					if ( dMin > hist.getLow().doubleValue() ) {
						dMin = hist.getLow().doubleValue();
					}
				}
			}
			fdt = getNextDate( fr, to, fdt );
			if ( fdt == null ) {
				break;
			}
		}
		return mapHist;
	}

	/**
	* ファイルから読み込んだ行を解析して、RateHistRosokuを返します。
	* @param	line	読み込んだ行の内容
	* @return	解析結果
	*/
	protected RateHistRosoku parseRate( String line ) throws ParseException {
		CmnDate dt;
		BigDecimal	l_rate, r_rate, mid;

		String[] tok = CmnString.getStringToken( line, "\t" );
		if( !tok[0].equals( pair.getCcy1().getCode() ) ||
			!tok[1].equals( pair.getCcy2().getCode() ) ||
			!tok[2].equals( term ))
		{
			return null;
		}

		dt = new CmnDate( tok[6], "yyyy-MM-dd HH:mm:ss" );
		l_rate = new BigDecimal( tok[3].trim() )
						.setScale( pair.getPoint(), BigDecimal.ROUND_DOWN );
		r_rate = new BigDecimal( tok[4].trim() )
						.setScale( pair.getPoint(), BigDecimal.ROUND_DOWN );
		mid = l_rate.add( r_rate )
					.divide( new BigDecimal( "2" ),
						pair.getPoint(), BigDecimal.ROUND_DOWN );
		RateHistRosoku hist = new RateHistRosoku();
		hist.setStartDate( dt );
		hist.setEndDate( dt );
		hist.setStart( mid );
		hist.setEnd( mid );
		hist.setHigh( mid );
		hist.setLow( mid );
		return hist;
	}

	/**
	* レートのキーを返します。
	* @param	rate	読み込んだレート
	* @return	レートのキー
	*/
	protected String getRateKey( RateHistRosoku rate ) {
		String key = rate.getStartDate().format( "yyyyMMddHHmmss" );
		return key;
	}

	/**
	* レートを集約するキーを返します。
	* @param	span	チャート種別
	* @param	rate	読み込んだレート
	* @return	集約するキー
	*/
	protected String getHistoryKey( String span, RateHistRosoku rate ) {
		String key = null;
		CmnDate dt = rate.getStartDate();
		if ( span.equals( "5MIN" )) {
			DecimalFormat fmt = new DecimalFormat( "00" );
			key = dt.format( "yyyyMMddHH" )
					+ fmt.format( ( dt.get( CmnDate.MINUTE ) / 5 ) * 5 );
		}
		return key;
	}

	protected CmnDate getNextDate( CmnDate fr, CmnDate to, CmnDate dt ) {
		CmnDate ret = (CmnDate )dt.clone();
		ret.add( CmnDate.DATE, 1 );
		if ( !ret.before( to )) {
			return null;
		}
		return ret;
	}
}
